#include "inc/main.h"

/* 
Krulowa Fal v 2.3
QuakeWorld server query and find player
UDP based daemon
*/

/* GLOBALS */

struct serversearchinfo *slist = NULL;
FILE *logs;
struct tm *nao;
time_t lt;

/* FUNCTIONS */

/*
StripColors (char *colortext) - translating quake color chars to plaintext
*/
char *StripColors(char *colortext) {

	int i = 0;
	int j = 18;
	static char plaintext[2048];
	memset(plaintext,0,2048);
	
	if (strlen(colortext) == 0 || strlen(colortext) > 2048) {
		return "";
	} else {		
		for (i = 0; i<strlen(colortext); i++) {
			// weird chars
			if (colortext[i] == -128) plaintext[i] = '<';
			// colored letters
			else if (colortext[i] & 128) {
				plaintext[i] = colortext[i] & 127;
			} else {
				plaintext[i] = colortext[i];
			}
			// numbers
			for (j = 18; j < 28; j++) {
				if (plaintext[i] == j) plaintext[i] += 30;
			}
			// fun chars
			switch (plaintext[i]) {
				case 2:
				case 13:
					plaintext[i] = '>';
					break;
				case 29:
				case 40:
					plaintext[i] = '(';
					break;
				case 31:
				case 41:
					plaintext[i] = ')';
					break;
				case 16:
				case 91:
					plaintext[i] = '[';
					break;
				case 17:
				case 93:
					plaintext[i] = ']';
					break;
				case 5:
				case 14:
				case 15:
				case 28:
				case 46:
					plaintext[i] = '.';
					break;
				case 33:
					plaintext[i] = '!';
					break;
				case 35:
					plaintext[i] = '#';
					break;
				case 37:
					plaintext[i] = '%';
					break;
				case 38:
					plaintext[i] = '&';
					break;
				case 39:
					plaintext[i] = '\\';
					break;
				case 43:
					plaintext[i] = '+';
					break;
				case 45:
					plaintext[i] = '-';
					break;
				case 64:
					plaintext[i] = '@';
					break;
				case 123:
					plaintext[i] = '{';
				case 125:
					plaintext[i] = '}';
			}
			// FIXME: are there any cases missing?
		}
	}
	strcat(plaintext,"\0");
	return plaintext;
}


/* 
char *StripChars(char *searchtext) - remove unwanted characters in the searchstring
*/
char *StripChars(char *searchtext) {

	char *ptr;
	int serlen;
	int i=0; // truncate the searchtext to 16 chars
	
	static char fixed[32];

	memset(fixed,0,32);
	ptr = searchtext;
	
	while (strlen(ptr) > 0 && i < 16) {
		if (*ptr != '\"' && *ptr != '\'') {
			strncat(fixed,ptr,1);
			*ptr++;
			i++;
		} else {
			*ptr++;
			i++;
		}
	}

	strcat(fixed,"\0");
	return fixed;
}

/*
char *ParseServerInfo(char *string, char *lookmeup) - get values for serverinfo
*/
char *ParseServerInfo(char *string, char *lookmeup) {

	char *p,*p1,*p2;
	static char my_return[128];
	memset(my_return,0,128);

	p = strstr(string,lookmeup);
	if (!p) {
		strcpy(my_return,"none");
		return my_return;
	}
	p += strlen(lookmeup);
	p1 = strchr(p,'\\');
	p2 = strchr(p,'\n');
	if (p1) {
		strncpy(my_return,p,p1-p);
	} else {
		if (p2) {
			strncpy(my_return,p,p2-p);
		} else {
			strncpy(my_return,p,strlen(p));
		}
	}
	return my_return;
}


/*
struct serverinfo *DealWithServerInfo(char *servstring) - fill in the serverinfo
*/
struct serverinfo *DealWithServerInfo(char *servstring) {

	struct serverinfo *serv;
	serv = malloc(sizeof(*serv));
	static char raw_hostname[128];
	memset(raw_hostname,0,128);

	serv->maxclients = atoi(ParseServerInfo(servstring,"\\maxclients\\"));
	serv->deathmatch = atoi(ParseServerInfo(servstring,"\\deathmatch\\"));
	strcpy(raw_hostname,ParseServerInfo(servstring,"\\hostname\\"));
	strcpy(serv->hostname,StripColors(raw_hostname));
	strcpy(serv->map,ParseServerInfo(servstring,"\\map\\"));
	strcpy(serv->status,ParseServerInfo(servstring,"\\status\\"));
	
	return serv;
}


/*
Add new players to a list
*/
void AddNewPlayer(int frags, char *name, char *team, struct playerinfo **plist) {

	struct playerinfo *newplayer;
	newplayer = malloc(sizeof(struct playerinfo));

	newplayer->frags = frags;
	
	memset(newplayer->nickname,0,sizeof(newplayer->nickname));
	strcpy(newplayer->nickname,name);
	strcat(newplayer->nickname,"\0");
	
	memset(newplayer->team,0,sizeof(newplayer->team));
	strcpy(newplayer->team,team);
	strcat(newplayer->team,"\0");

	newplayer->next = NULL;

	if (!plist) {
		*plist = newplayer;
	} else {
		newplayer->next = *plist;
		*plist = newplayer;
	}
}


/*
Add new servers to the list
*/
void AddNewServer(int ID, char IP[16], unsigned short port, char country[3]) {

	struct serversearchinfo *newserver;
	newserver = malloc(sizeof(struct serversearchinfo));

	newserver->ID = ID;
	strcpy(newserver->IP,IP);
	newserver->port = port;
	strcpy(newserver->country,country);
	newserver->next = NULL;

	if (!slist) {
		slist = newserver;
	} else {
		newserver->next = slist;
		slist = newserver;
	}
}


/* 
Parse playerinfo data to a struct and add to player list
*/
int DealWithPlayerInfo(char *playstring, struct playerinfo **plist) {

	char	name[64];
	char	team[32];
	int	frags, len, rc, top, bottom;
	char	*cur, *end;

	len = strlen(playstring);
	cur = playstring;
	end = playstring;
	int playernum = 0;
	
	if ((strstr(playstring,"\"\n")) != NULL) {
		// response with teamtags from newer mvdsv servers
		do {
			memset(name,0,64);
			memset(team,0,32);
			frags = 0;
			if ((strlen(end)) < 2 ) break;

			/* frags */
			rc = sscanf(cur,"%*d %d %*d %*d",&frags);

			/* nickname */
			cur = strchr(cur,'"');
			cur += 1;
			end = strchr(cur,'"');
			memset(name,0,64);
			strncpy(name,cur,end-cur);
			
			/* team */
			cur = strstr(end,"\" \"");
			cur += 3;
			cur = strchr(cur,'"');
			cur += 2;
			cur = strchr(cur,'"');
			cur += 1;
			end = strchr(cur,'"');
			if ((strlen(cur)) <= (strlen(end))) {
				memset(team,0,32);
			} else {
				memset(team,0,32);
				strncpy(team,cur,end-cur);
			}

			end += 2; // next line
			cur = end;
			if (frags < 0 || frags > 666) frags = -9999;
			if (frags != -9999) playernum++;
			strcpy(name,StripColors(name));
			strcpy(team,StripColors(team));

			AddNewPlayer(frags,name,team,plist);
		
		} while (strlen(end) <= len);
	} else {
		do {
		// legacy stuff for old mvdsv - no teamtags
			memset(name,0,64);
			memset(team,0,32);
			if ((strlen(end)) < 2 ) break;

			/* frags */
			rc = sscanf(cur,"%*d %d %*d %*d",&frags);

			/* nickname */
			cur = strchr(cur,'"');
			cur += 1;
			end = strchr(cur,'"');
			strncpy(name,cur,end-cur);
			strcat(name,"\0");

			/* team - unused for now */
			end += 1;
			cur = strstr(end,"\" ");
			cur += 2;
			rc = sscanf(cur,"%d %d",&top,&bottom);
			end = strchr(cur,'\n');
			snprintf(team,sizeof(top)+sizeof(bottom)+1,"%i_%i",top,bottom);
			strcat(name,"\0");

			cur = end;
			end += 1;

			if (frags < 0 || frags > 666) frags = -9999;		
			if (frags != -9999) playernum++;
			strcpy(name,StripColors(name));
			strcpy(team,StripColors(team));
			
			AddNewPlayer(frags,name,team,plist);

		} while (strlen(end) <= len);
	}
	return playernum;
}


/*
ShowPlayer* - print info about players on current server in proper mode
*/
char *ShowPlayerListDuel(struct playerinfo *playerlist) {

	static char reply[256];
	char treply[256];
	int suc = 0;
	
	memset(reply,0,256);
	memset(treply,0,256);

	while (playerlist) {
		if (playerlist->frags == -9999) {
			playerlist = playerlist->next;
			continue;
		} else { 
			if (!suc) {
				memset(treply,0,sizeof(treply));
				sprintf(treply,"DUEL: %s [%d] VS ",playerlist->nickname,playerlist->frags);
				strcat(reply,treply);
				playerlist = playerlist->next;
				suc = 1;
				continue;
			} else {
				memset(treply,0,sizeof(treply));			
				sprintf(treply,"[%d] %s\n",playerlist->frags,playerlist->nickname);
				strcat(reply,treply);
				playerlist = playerlist->next;
			}
		}
	}
	strcat(reply,"\0");
	return reply;
}


char *ShowPlayerListOther(struct playerinfo *playerlist) {

	static char reply[512];
	char treply[32];
	memset(reply,0,512);
	memset(treply,0,32);
	char stringend[2];
	strcpy(stringend,",\0");

	struct playerinfo *templist = playerlist;
	
	while (templist) {
		if (templist->frags == -9999) {
			templist = templist->next;
		} else {
			if (templist->next == NULL) strcpy(stringend,".\0");
			memset(treply,0,sizeof(treply));
			sprintf(treply,"%s(%d)%s ",templist->nickname,templist->frags,stringend);
			strcat(reply,treply);
			templist = templist->next;
		}
	}
	strcat(reply,"\n");
	return reply;
}


char *ShowPlayerListTeam(struct playerinfo *playerlist) {
	
	char teama[8], teamb[8];
	static char reply[512];
	char treply[64];
	memset(reply,0,512);
	memset(treply,0,64);
	char stringend[2];
	
	int teamfrags = 0;
	struct playerinfo *templist = playerlist;

	/* get the teamnames */
	while (templist) {
		if (templist->frags != -9999) {
			printf("%s\n",templist->team);
			strcpy(teama,templist->team);
			strcat(teama,"\0");
			break;
		} else {
			templist = templist->next;
		}
	}
	while (templist) {
		if (templist->frags != -9999 && (strcmp(teama,templist->team)) != 0) {
			strcpy(teamb,templist->team);
			strcat(teamb,"\0");
			break;
		} else {
			templist = templist->next;
		}
	}

	/* print a line with results for the first team */
	templist = playerlist;
	sprintf(reply,"TEAM [%s]: ",teama);
	strcpy(stringend,",\0");

	while (templist) {
		if (templist->frags == -9999 || (strcmp(teama,templist->team)) != 0) {
			templist = templist->next;
		} else {
			if (templist->next == NULL) strcpy(stringend,".\0");
			memset(treply,0,sizeof(treply));
			sprintf(treply,"%s(%d)%s ",templist->nickname,templist->frags,stringend);
			strcat(reply,treply);
			teamfrags += templist->frags;
			templist = templist->next;
		}
	}
	memset(treply,0,sizeof(treply));
	sprintf(treply,"TOTAL: %d\n",teamfrags);
	strcat(reply,treply);
	
	/* print a line with results for the other team */
	templist = playerlist;
	teamfrags = 0;
	sprintf(treply,"TEAM [%s]: ",teamb);
	strcat(reply,treply);
	strcpy(stringend,",\0");

	while (templist) {
		if (templist->frags == -9999 || (strcmp(teamb,templist->team)) != 0) {
			templist = templist->next;
		} else {
			if (templist->next == NULL) strcpy(stringend,".\0");
			memset(treply,0,sizeof(treply));
			sprintf(treply,"%s(%d)%s ",templist->nickname,templist->frags,stringend);
			strcat(reply,treply);
			teamfrags += templist->frags;
			templist = templist->next;
		}
	}
	memset(treply,0,sizeof(treply));
	sprintf(treply,"TOTAL: %d\n",teamfrags);
	strcat(reply,treply);
	return reply;
}


/*
char *SendQuery(char *ServerIP, unsigned short ServerPort) - Send server 
query and return the received data for parsing
*/
char *SendQuery(char *ServerIP, unsigned short ServerPort) {
	
	ssize_t recv_ret;
	int MySocket;
	struct sockaddr_in ServerAddress;
	struct sockaddr_in ReplyAddress;
	unsigned int ReplyLen;
	static char buffer[2048];
	
	fd_set MySet;
	FD_ZERO(&MySet);

	struct timeval tmout;
	tmout.tv_sec = 1;
	tmout.tv_usec = 0;

	static const char senddata[] = {255, 255, 255, 255, 's','t','a','t','u','s',' ','2','3','\n'};
	size_t datalen = sizeof(senddata);
	
	if ((MySocket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
		perror("socket()"); 
	}
	
	

	memset(&ServerAddress, 0, sizeof(ServerAddress));
	ServerAddress.sin_family = AF_INET;
	ServerAddress.sin_addr.s_addr = inet_addr(ServerIP);
	ServerAddress.sin_port = htons(ServerPort);
	
	if (sendto(MySocket, senddata, datalen, 0, (struct sockaddr *)&ServerAddress, sizeof(ServerAddress)) == -1) { 
		perror("sendto()");
	}
	ReplyLen = sizeof(ReplyAddress);

	FD_SET(MySocket,&MySet);

	if (FD_ISSET(MySocket,&MySet)) {
		if (select(MySocket+1, &MySet, NULL, NULL, &tmout) == 0) {
			sprintf(buffer,"The server did not respond or it took to long.\nGUWNOGG");
			return buffer;
		} else {
			recvfrom(MySocket, &buffer, sizeof(buffer), 0, (struct sockaddr *)&ReplyAddress, &ReplyLen);
		}
	}

	if (ServerAddress.sin_addr.s_addr != ReplyAddress.sin_addr.s_addr) {
		printf("Unexpected packet.\n");
	}
	
	close(MySocket);
	return buffer;
}


/* 
char *SendMassQuery(char *searchtext, char country[3]) - Send queries to all 
servers in the list and
*/
char *SendMassQuery(char searchtext[32], char country[3]) {
	
	int MySocket;
	
	struct timeval tmout;
	fd_set MySet;
	FD_ZERO(&MySet);

	struct serversearchinfo *tmp = slist;

	struct sockaddr_in ServerAddress;
	struct sockaddr_in ReplyAddress;
	unsigned int ReplyLen;
	ReplyLen = sizeof(ReplyAddress);

	static char buffer[2048];
	static char cleanbuffer[2048];
	static char repldata[2048];
	char temprepl[128];
	memset(repldata,0,2048);
	memset(temprepl,0,128);

	char *hit;
	int servernumber;
	int servercounter = 0;

	char csearchtext[32];
	memset(csearchtext,0,32);
	strcpy(csearchtext,StripChars(searchtext));
	
	static const char senddata[] = {255, 255, 255, 255, 's','t','a','t','u','s',' ','2','2','\n'}; // only playerinfo
	size_t datalen = sizeof(senddata);
	int success = 0;

	int parts = 0;

	while (parts < 6) {
		parts++;
		servernumber = 0;
		if ((MySocket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
			perror("socket()"); 
		}
		if (fcntl(MySocket,F_SETFL,O_NONBLOCK|O_ASYNC) < 0) {
			perror("fcntl()");
		}

		while (tmp) {
			if ((strcasecmp(country,"xz")) == 0) {
				memset(&ServerAddress, 0, sizeof(ServerAddress));
				ServerAddress.sin_family = AF_INET;
				ServerAddress.sin_addr.s_addr = inet_addr(tmp->IP);
				ServerAddress.sin_port = htons(tmp->port);
				sendto(MySocket, senddata, datalen, 0, (struct sockaddr *)&ServerAddress, sizeof(ServerAddress));
				tmp = tmp->next;
			} else {
				if ((strcasecmp(tmp->country,country)) == 0) {
					memset(&ServerAddress, 0, sizeof(ServerAddress));
					ServerAddress.sin_family = AF_INET;
					ServerAddress.sin_addr.s_addr = inet_addr(tmp->IP);
					ServerAddress.sin_port = htons(tmp->port);
					sendto(MySocket, senddata, datalen, 0, (struct sockaddr *)&ServerAddress, sizeof(ServerAddress));
					tmp = tmp->next;
				} else {
					tmp = tmp->next;
				}
			}
			if (servernumber == 35) { // get info in parts of 35
				break;
			} else {
				servernumber++;
			}
		}

		tmout.tv_sec = 1;
		tmout.tv_usec = 0; 
		FD_SET(MySocket,&MySet);
		int mylen = 0;
		struct playerinfo *players;
		players = NULL;
		int retn;

		if (FD_ISSET(MySocket,&MySet)) {
			memset(buffer,0,sizeof(buffer));
			while (select(MySocket+1, &MySet, NULL, NULL, &tmout) > 0) {
				recvfrom(MySocket, &buffer, sizeof(buffer), 0, (struct sockaddr *) &ReplyAddress, &ReplyLen);
				servercounter++;
				if ((mylen = strlen(buffer)) < 10 || (hit = strchr(buffer,'"')) == NULL) continue;	// short info or server info without players
				printf("IP %s:%i \nbuffer: %s\n",inet_ntoa(ReplyAddress.sin_addr),ntohs(ReplyAddress.sin_port),buffer);
				memset(cleanbuffer,0,2048);
				strcpy(cleanbuffer,StripColors(buffer));						// buffer has gay nicknames inside, STRIP!
				retn = DealWithPlayerInfo(cleanbuffer,&players);
				while (players) {
					if ((hit = strcasestr(players->nickname,csearchtext)) != NULL) {
						sprintf(temprepl,"%s:%d - %s\n",inet_ntoa(ReplyAddress.sin_addr),ntohs(ReplyAddress.sin_port),players->nickname);
						strcat(repldata,temprepl);
						memset(temprepl,0,128);
						success = 1;
					} 
					free(players);
					players = players->next;
				}
				players = NULL;
				memset(buffer,0,sizeof(buffer));
			}
		}
		FD_ZERO(&MySet);
		close(MySocket);
	}
	if (!success) {
		sprintf(repldata,"%s was not found on any of the %i scanned servers.\n",csearchtext,servercounter);
	} else {
		sprintf(temprepl,"(%i scanned servers)\n",servercounter);
		strcat(repldata,temprepl);
	}

	memset(buffer,0,2048);		
	return repldata;
}


/*
char *ProcessReply (char *reply) - process the query reply and generate answer
*/
char *ProcessReply (char *reply) {
	
	static char repldata[4096];
	memset(repldata,0,4096);

	int playernum = 0;
	
	int replen;

	if ((replen = strlen(reply)) < 6) {
		strcpy(repldata,"The server speaks chinese, kthxbai.\0");
		return repldata;
	}
	
	struct playerinfo *plist = NULL;
	char *d;
	char servstring[2048];
	char playstring[2048];
	if ((d=strstr(reply,"GUWNOGG")) != NULL) return strncpy(repldata,reply,d-reply);
	d = strstr(reply,"\n");
	strncpy(servstring,reply,d-reply+1);
	strncpy(playstring,d+1,strlen(reply));
	
	char version[32];
	strcpy(version,ParseServerInfo(servstring,"qwe_version\\")); // need this to deal with shitty mvdsv 0.19.20

	if (strlen(playstring) > 3 ) {
		playernum = DealWithPlayerInfo(playstring,&plist);
	}

	// serverinfo
	struct serverinfo *a;	
	a = DealWithServerInfo(servstring);
	if (a->hostname[strlen(a->hostname)-1] == -121) a->hostname[strlen(a->hostname)-1] = 0; // clear the weird 
												//char in game mode
	a->playerlist = plist;
	
	if (strstr(a->hostname,"left") != NULL) {
		sprintf(repldata,"SERVER: %s, %d/%d, %s, dmm%d, STATUS: %s\n",StripColors(a->hostname),playernum,a->maxclients,a->map,a->deathmatch,a->status);
	} else {
		sprintf(repldata,"SERVER: %s, %d/%d, %s, dmm%d, STATUS: %s\n",a->hostname,playernum,a->maxclients,a->map,a->deathmatch,a->status);
	}		
	// playerlist
	if (playernum != 0) {
		if ((playernum % 2) != 0 || (strcmp(a->status,"none") == 0) || (strcmp(a->status,"Standby") == 0)|| strstr(version,"0.19")) { // FFA or game with uneven number of players OR OLD SHIT FFS
			strcat(repldata,ShowPlayerListOther(a->playerlist));
		} else if (playernum == 2) { // normal duel
			strcat(repldata,ShowPlayerListDuel(a->playerlist));
		} else if (playernum == 4 || playernum == 6 || playernum == 8) { // 2on2/3on3/4on4 tdm
			strcat(repldata,ShowPlayerListTeam(a->playerlist));
		}
	}
	playernum = 0;
	while (plist) { free(plist); plist = plist->next;}
	plist = NULL;
	return repldata;
}


/*
* Read and parse information from XML server base
*/
static void LoadServers(const char *filename) {
	
	xmlTextReaderPtr reader;
	int ret;

	int ID = 0;
	char IP[16];
	unsigned short port = 0;
	char country[3];
	const xmlChar *value;
	reader = xmlReaderForFile(filename, NULL, 0);
	if (reader != NULL) {
		ret = xmlTextReaderRead(reader);
		ret = xmlTextReaderRead(reader);
		ret = xmlTextReaderRead(reader);
		while (ret == 1) {
			if (xmlTextReaderDepth(reader) == 1 && xmlTextReaderNodeType(reader) == 1) { // <server> tag
				/* ID */
				ret = xmlTextReaderRead(reader);
				ret = xmlTextReaderRead(reader);
				ret = xmlTextReaderRead(reader);
				value = xmlTextReaderConstValue(reader);
				ID = atoi(value);
				/* IP */
				ret = xmlTextReaderRead(reader);
				ret = xmlTextReaderRead(reader);
				ret = xmlTextReaderRead(reader);
				ret = xmlTextReaderRead(reader);
				value = xmlTextReaderConstValue(reader);
				strcpy(IP,value);
				/* PORT */
				ret = xmlTextReaderRead(reader);
				ret = xmlTextReaderRead(reader);
				ret = xmlTextReaderRead(reader);
				ret = xmlTextReaderRead(reader);
				value = xmlTextReaderConstValue(reader);
				port = atoi(value);
				/* COUNTRY */				
				ret = xmlTextReaderRead(reader);
				ret = xmlTextReaderRead(reader);
				ret = xmlTextReaderRead(reader);
				ret = xmlTextReaderRead(reader);
				value = xmlTextReaderConstValue(reader);
				strcpy(country,value);
				AddNewServer(ID,IP,port,country);				
				ret = xmlTextReaderRead(reader);
			} else {
				ret = xmlTextReaderRead(reader);
			}
		}
		xmlFreeTextReader(reader);
	} else {
		fprintf(logs,"[%2i-%2i %2i:%2i] Refreshing the server list: FAIL - XML Reader failed.\n",nao->tm_mday,nao->tm_mon,nao->tm_hour,nao->tm_min);
		fflush(logs);
	}
}

/*
char *ProcessRequest(char request[128]) - read requests and decide what to do
*/
char *ProcessRequest(char request[128]) {

	char *ServerIP;
	char port[6],ctr[3];
	int ServerPort,a;
	struct hostent *myhost;
	struct in_addr in;
	char cmd[3], addr[64], param[32], failure[128];
	char *left,*right;
	char *reply;
	int scanned = 0;
	int len;

	memset(cmd,0,3);
	memset(addr,0,64);
	memset(param,0,32);
	memset(port,0,6);
	
	/* get command, default is query */
	left = request;
	if ((right = strchr(request,' ')) != NULL) {
		strncpy(cmd,left,right-left);
		strcat(cmd,"\0");
		right+=1;
		left = right;
	} else {
		strcpy(cmd,"q");
	}
	if (strcasecmp(cmd,"q") == 0) {
		/* get address */
		if ((right = strchr(request,':')) == NULL) {
			strcpy(addr,left);
			strcat(addr,"\0");
			ServerPort = 27500;
		} else {
			strncpy(addr,left,right-left);
			strcpy(port,++right);
			strcat(port,"\0");
			ServerPort = atoi(port);
		}
		/* hostname or IP */
		if (scanned = sscanf(addr,"%d.%d.%d.%d",&a,&a,&a,&a) == 4) {
			reply = SendQuery(addr,ServerPort);
			reply = ProcessReply(reply);
		} else {
			struct hostent *myhost;
			struct in_addr in;
			if ((myhost = gethostbyname(addr)) == NULL) {
				strcpy(failure,"That host is invalid.\0");
				reply = failure;
				return reply;
			}
			if (myhost->h_addr_list[0] != NULL) {
				memcpy(&in.s_addr, myhost->h_addr_list[0], sizeof(in.s_addr));
				ServerIP = inet_ntoa(in);
				reply = SendQuery(ServerIP,ServerPort);
				reply = ProcessReply(reply);
			} else {
				strcpy(failure,"That host is invalid.\0");
				reply = failure;
				return reply;
			}
		}
	} else if (strcasecmp(cmd,"fp") == 0) {
		if ((right = strchr(right,' ')) != NULL) { // there is country param
			strncpy(param,left,right-left);
			strcat(param,"\0");
			strcpy(ctr,++right);
			strcat(ctr,"\0");
		} else {
			strcpy(param,left);
			strcat(param,"\0");
			strcpy(ctr,"xz\0");
		}
		if ((len = strlen(param)) <= 1) {
			sprintf(failure,"Usage: !qw fp <nickname> [country]\n");
			reply = failure;
			return reply;
		}
		reply = SendMassQuery(param,ctr);
		xmlCleanupParser();
	} else {
		strcpy(failure,"That host is invalid.\0");
		reply = failure;
		return reply;
	}
	return reply;
}


/*
static size_t write_data(void *ptr, size_t size, size_t nmemb, void *stream) - file writer function for curl
*/
static size_t write_data(void *ptr, size_t size, size_t nmemb, void *stream) {
	int written = fwrite(ptr, size, nmemb, (FILE *)stream);
	return written;
}


int RefreshServerListFile() {
	CURL *curl_handle;
	static const char *bodyfilename = "servers.temp";
	FILE *bodyfile;

	curl_global_init(CURL_GLOBAL_ALL);
	curl_handle = curl_easy_init();
	curl_easy_setopt(curl_handle, CURLOPT_URL, "http://www.quakeservers.net/european_servers.php");
	curl_easy_setopt(curl_handle, CURLOPT_NOPROGRESS, 1);
	curl_easy_setopt(curl_handle, CURLOPT_WRITEFUNCTION, write_data);

	if ((bodyfile = fopen(bodyfilename,"w")) == NULL) {
		curl_easy_cleanup(curl_handle);
		return -1;
	} else {
		curl_easy_setopt(curl_handle, CURLOPT_WRITEDATA, bodyfile);
		curl_easy_perform(curl_handle);
		fclose(bodyfile);
		curl_easy_cleanup(curl_handle);
		return 0;
	}
}

void RefreshServers() {
	struct stat servlistfile;
	int refreturn;

	if ((refreturn = RefreshServerListFile()) == 0) {
		if (stat("servers.temp", &servlistfile) != 0) {
			fprintf(logs,"[%2i-%2i %2i:%2i] Refreshing the server list: FAIL - could not stat servers.temp.\n",nao->tm_mday,nao->tm_mon,nao->tm_hour,nao->tm_min);
			fflush(logs);
		} else {
			if (servlistfile.st_size < 4096) {
				fprintf(logs,"[%2i-%2i %2i:%2i] Refreshing the server list: FAIL - server list incomplete.\n",nao->tm_mday,nao->tm_mon,nao->tm_hour,nao->tm_min);
				fflush(logs);
			} else {
				if ((rename("servers.temp", "servers.xml")) < 0) {
					fprintf(logs,"[%2i-%2i %2i:%2i] Refreshing the server list: FAIL - could not move to servers.xml.\n",nao->tm_mday,nao->tm_mon,nao->tm_hour,nao->tm_min);
					fflush(logs);
				} else {								// downloaded a file and it contains a valid server list
					while (slist) {
						free(slist); 
						slist = slist->next;
					}
					slist = NULL;
					LoadServers("servers.xml");
					fprintf(logs,"[%2i-%2i %2i:%2i] Refreshing the server list: SUCCESS.\n",nao->tm_mday,nao->tm_mon,nao->tm_hour,nao->tm_min);
					fflush(logs);
				}
			}
		}
	} else {
		fprintf(logs,"[%2i-%2i %2i:%2i] Refreshing the server list: FAIL - CURL failed.\n",nao->tm_mday,nao->tm_mon,nao->tm_hour,nao->tm_min);
		fflush(logs);
	}		
}

/*
MAIN: get packet, send reply
*/
int main(int argc, char *argv[]) {
	int sock, length, fromlen, n;
	struct sockaddr_in server;
	struct sockaddr_in from;
	char buf[128];
	char reply[4096];
	lt = time(NULL);
	nao = localtime(&lt);

	if ((logs = fopen("kf.log","a+")) == NULL) {
		perror("fopen");
		exit(0);
	}

	if (argc < 2) {
		fprintf(stderr, "Usage: %s <port>\n",argv[0]);
		exit(0);
	}

	sock=socket(AF_INET, SOCK_DGRAM, 0);

	if (sock < 0) {
		fprintf(stderr, "Cannot create socket, aborting.\n");
		exit(0);
	}
	
	if (fcntl(sock,F_SETFL,O_NONBLOCK|O_ASYNC) < 0) {
		perror("fcntl()");
		exit(0);
	}

	length = sizeof(server);
	bzero(&server,length);
	server.sin_family=AF_INET;
	server.sin_addr.s_addr=INADDR_ANY;
	server.sin_port=htons(atoi(argv[1]));
	
	if (bind(sock,(struct sockaddr *)&server,length)<0) {
		fprintf(stderr, "Cannot bind(), aborting (check port).\n");
		exit(0);
	}

	fromlen = sizeof(struct sockaddr_in);

	RefreshServers();

	daemon(1,0);

	while (1) {
		sleep(1);
		lt = time(NULL);
		nao = localtime(&lt);
		if ((nao->tm_sec == 1) && (nao->tm_min == 05 || nao->tm_min == 38)) RefreshServers();
		memset(buf,0,128);

		if ((n = recvfrom(sock,buf,128,0,(struct sockaddr *)&from,&fromlen)) > 0) { 
			fprintf(logs,"[%2i-%2i %2i:%2i] Received %s\n",nao->tm_mday,nao->tm_mon,nao->tm_hour,nao->tm_min,buf);
			fflush(logs);
			memset(reply,0,4096);
			strcpy(reply,ProcessRequest(buf));
			if ((n = sendto(sock,reply,sizeof(reply),0,(struct sockaddr *)&from,fromlen)) < 0) { 
				perror("sendto");
			}
		}
	}
	fclose(logs);
}
